{% extends "!layout.html" %}

{%- block extrahead %}
{{ super() }}

    <META HTTP-EQUIV="CACHE-CONTROL" CONTENT="NO-CACHE">
    <META HTTP-EQUIV="PRAGMA" CONTENT="NO-CACHE">

    <!-- Point to the *latest* version switcher. This will allow the latest versions to appear on older documentation. -->
    <script type="text/javascript" src="https://scitools.org.uk/cartopy/docs/latest/_static/version_switch.js"></script>
{% endblock %}
